# How to make my Airflow schedule tasks on kubernetes cluster?
Usually the airflow machine is spawning new tasks as containers on the current VM where airflow itself is running. Unfortunately this exposes the VM to be saturated by the jobs it schedules and it is not scalable at all.

A solution can be: spawn new containers in the Kubernetes cluster directly via the kubectl command and off-load the jobs to the cluster.

# Requirements

1. Install Openstack packages:
    ```
    pip install python-openstackclient
    pip install python-magnumclient
    ```
1. Create an executable script **start_developer_project.sh** so that once executed will connect you to the project containing your cluster. It contains the openstack code that you can download on the top right corner of the Openstack web page. Tools > OpenStack RC v3.
    ```
    vim start_developer_project.sh
    (past the content of OpenStack RC v3 save and close)
    chmod 711 start_developer_project.sh
    ```
1. To access the project containing the clusters call the following:
    ```
    source start_developer_project.sh
    ```
    Check by trying this command to see the list of all clusters in the project:
    ```
    openstack coe cluster list
    ```
1. Create the config file for accessing your specific cluster with kubectl:
    ```
    openstack coe cluster config cluster-name
    export KUBECONFIG=<absolute path to your newly generated config file>
    ```
    Check by trying this command to see all resources in the cluster (you should see the one there by default in every cluster):
    ```
    kubectl get all -A
    ```
1. Install secrets with yaml file **user_secret.yaml**
    ```
    apiVersion: v1
    kind: Secret
    metadata:
      name: my-kinit-secret
    type: kubernetes.io/basic-auth
    stringData:
      username: your_kinit_username
      password: your_kinit_pass
    ```
1. Install fluentd (make sure it is connected to inject data analytics config)
    ```
    helm install fluentd cern/fluentd --namespace kube-system  --values values.yaml
    ```
**values.yaml** content:
```
image:
  name: gitlab-registry.cern.ch/cloud/atomic-system-containers/fluentd-elasticsearch
  tag: v2.4.0
plugins:
  - fluent-plugin-rewrite-tag-filter
  - fluent-plugin-out-http
  - fluent-plugin-grok-parser
output:
  producer: something
  endpoint: something
  includeInternal: false
# Implies that docker is used and produces logs in the json-file format.
# https://docs.docker.com/config/containers/logging/json-file/
containerRuntime: "host-docker"
parsingConfig: |

    <source>
      @type forward
      port 24224
      bind 0.0.0.0
    </source>
    # Relabel data https://docs.fluentd.org/output/relabel
    <match data-analytics.*>
      @type copy
      <store>
        # for the data analytics results
        @type relabel
        @label @analytics_results
      </store>
      <store>
        # for the rest
        @type relabel
        @label @any_printout
      </store>
    </match>
    <label @analytics_results>
      <filter data-analytics.*>
      # Find only the printouts with key results
        @type grep
        <regexp>
          key log
          pattern ^\{\"results\".*$
        </regexp>
      </filter>
      <filter data-analytics.*>
      # Rename log key -> log: {} -> results: {}
        @type parser
        key_name log
        reserve_data true
        remove_key_name_field true
        <parse>
          @type json
        </parse>
      </filter>
      # Move all data in the 'data' field as expected by http://monitdocs.web.cern.ch/monitdocs/logs/service_logs.html
      <filter data-analytics.*>
      # https://docs.fluentd.org/filter/record_transformer
        @type record_transformer
        #https://github.com/repeatedly/fluent-plugin-record-modifier/issues/39
        enable_ruby true
        auto_typecast true
        remove_keys container_id
        <record>
          data ${record}
        </record>
        renew_record true
        keep_keys data
      </filter>
      # Add metadata as expected by MONIT http://monitdocs.web.cern.ch/monitdocs/logs/service_logs.html
      <filter data-analytics.*>
      # https://docs.fluentd.org/filter/record_transformer
        @type record_transformer
        enable_ruby true
        <record>
          producer "openstack"
          type_prefix "raw"
          type "analytics"
          _id ${record["data"]["results"]["document_id"]}
          another_id ${record["data"]["results"]["document_id"]}
          timestamp ${record["data"]["results"]["temporal"]["time_end_milli"]}
          timestamp_format "yyyy-MM"
        </record>
      </filter>
      <match data-analytics.*>
        @type copy
        # send data to http endpoint
        <store>
          @type http
          #url http://srv_monit_fluentd:9000/data-analytics
          endpoint_url http://monit-metrics.cern.ch:10012
          open_timeout 2
          http_method post
          <format>
            @type json
          </format>
          <buffer>
            flush_interval 10s
          </buffer>
        </store>
        # send data to stdout
        <store>
          @type stdout
          <buffer tag>
            flush_mode immediate
            #flush_interval 1s
            flush_at_shutdown true
            delayed_commit_timeout 1s
          </buffer>
        </store>
      </match>
    </label>
    <label @any_printout>
      <match *.*>
        @type copy
        <store>
          @type stdout
          <buffer tag>
            flush_mode immediate
            #flush_interval 1s
            flush_at_shutdown true
            delayed_commit_timeout 1s
          </buffer>
        </store>
      <store>
        # as from https://knplabs.com/en/blog/how2tips-how-to-test-fluentd-config
        @type file
        path /tmp/myfluent.log
              append true
              add_path_suffix false
              <format>
                  @type json
              </format>
              <inject>
                  tag_key tag
              </inject>
              <buffer []>

                  @type memory
                  chunk_limit_records 1
                  retry_max_times 0
              </buffer>
      </store>
      </match>
    </label>
    ```
1. Install eos
    ```
    helm install myeos cern/eosxd
    ```

## Monitoring
